/*
** calc.c
**
** Part of RPN calculator
** Copyright (c) 2005, 2006, 2007, 2010 NuukiaWorld
**
**  RPN calculator is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* includes nokix headers */
#include <nokix/all.h>
#include <math.h>
#include "float.h"
#define  INT_MAX  0x7FFFFFFF

#include "main.h"
#include "draw.h"
#include "stack.h"

/*******************************************************/

extern   double gamma1p( double x );

/*******************************************************/

#define  ONE_PLUS_EPSILON ( 1.0 + DBL_EPSILON )

/*******************************************************/

bool_t check_value( double d ) {

   if( fabs(d) <= DBL_MAX ) return true;
   display_msg( DMI_INFORMATION, &str_range );
   return false;
}

/* Calculate Greatest Common Denominator
*/
static int gcd( int a, int b ) {

   int t;

   while( b != 0 ) {
      t = b;
      b = a % b;
      a = t;
   }
   return a;
}

/* Converts decimal value to format H.MMSSs
*/
static double to_hms( double d ) {

   double h, m, s;
   
   d *= ONE_PLUS_EPSILON;  // to avoid problems caused by inaccurate numbers
   d = modf( d, &h ) * 60;
   s = modf( d, &m ) * 60;
   return h + m / 100 + s / 10000;
}

/* Converts H.MMSSs value to decimal
*/
static double from_hms( double d ) {

   double h, m, s;

   d *= ONE_PLUS_EPSILON;  // to avoid problems caused by inaccurate numbers
   d = modf( d, &h ) * 100;
   s = modf( d, &m ) * 100;
   return h + m / 60 + s / 3600;
}

/* Converts DD.MMYYYY value to a datetime structure
*/
static bool_t to_datetime( double d, struct datetime *dt ) {

   double dd, dm, dy;
   unsigned int id, im, iy;
   bool_t b;

   d *= ONE_PLUS_EPSILON;  // to avoid problems caused by inaccurate numbers
   d = modf( d, &dd ) * 100;
   dy = modf( d, &dm ) * 10000;
   dt->day = id = dd;
   dt->month = im = dm;
   dt->year = iy = dy;
   dt->hours = dt->mins = dt->secs = 0;
   
   // Check the validity of the date
   add_datetime( dt, 0 );
   b = ( dt->day == id && dt->month == im && dt->year == iy );
   if(!b) display_msg( DMI_INFORMATION, &str_invalid_date );
   return b;
}

/*static bool_t check_days( int i ) {

   if( abs(i) <= INT_MAX / 86400 ) return true;
   // Too large number of days to add/subtract
   display_msg( DMI_INFORMATION, &str_range );
   return false;
}*/

/* Calculates an operator (=two operands), operator number
   0... in msg_argv[0]
*/
void calc_operator (void) {

	double d;
   struct datetime dt;
   int i, ts1, ts2;

	if (!stack_arg_check (2)) return;

	switch(msg_argv[0]) {
		case 0:
			d = stack->stack[1] + stack->stack[0];		// Add
			break;
		case 1:
			d = stack->stack[1] - stack->stack[0];		// Subtract
			break;
		case 2:
			d = stack->stack[1] * stack->stack[0];		// Multiply
			break;
		case 3:
			d = stack->stack[1] / stack->stack[0];      // Divide
			break;
      case 4:
         d = stack->stack[1] * ( 1 + stack->stack[0] / 100.0 );  // Change by %
			break;
      case 5:
         d = stack->stack[1] * ( stack->stack[0] / 100.0 );    // %
			break;
		case 6:
         d = ( stack->stack[0] / stack->stack[1] - 1.0 ) * 100.0;   // %ch
         break;
      case 7:
         d = stack->stack[0] / stack->stack[1] * 100.0;    // %t
         break;
      case 8:
         d = fmod( stack->stack[1], stack->stack[0] );     // Remainder
			break;
      case 9:
         d = (double)gcd( lround(stack->stack[1]), lround(stack->stack[0]) );   // GCD
         break;
      case 10:
         d = (double)( lround(stack->stack[1]) / gcd( lround(stack->stack[1]), lround(stack->stack[0]) )
            * lround(stack->stack[0]) );            // LCM
         break;
		case 11:
			d = pow (stack->stack[1], stack->stack[0]);	      // Power
			break;
		case 12:
			d = pow (stack->stack[1], 1.0 / stack->stack[0]); // Root
			break;
		case 13:
			d = (double)( lround(stack->stack[1]) & lround(stack->stack[0]) );	// And
			break;
		case 14:
			d = (double)( lround(stack->stack[1]) | lround(stack->stack[0]) );	// Or
			break;
		case 15:
			d = (double)( lround(stack->stack[1]) ^ lround(stack->stack[0]) );	// Xor
			break;
		case 16:
			d = (double)(int)( (unsigned int)lround(stack->stack[1]) << lround(stack->stack[0]) );  // Shift left
			break;
		case 17:
			d = (double)(int)( (unsigned int)lround(stack->stack[1]) >> lround(stack->stack[0]) );  // Shift right
			break;
		case 18:
         d = (double)( lround(stack->stack[1]) >> lround(stack->stack[0]) );    // Arithmetic shift right
			break;
		case 19:
			d = gamma1p(stack->stack[1]) / ( gamma1p(stack->stack[0]) * gamma1p( stack->stack[1] - stack->stack[0] ) ); // Combination
			break;
		case 20:
			d = gamma1p(stack->stack[1]) / gamma1p( stack->stack[1] - stack->stack[0] );  // Permutation
			break;
		case 21:
         if( !to_datetime( stack->stack[1], &dt ) ) return;              // Date add
         i = lround(stack->stack[0]);
//         if( !check_days(i) ) return;
         if( abs(i) > INT_MAX / 86400 ) {
            d = DBL_MAX + DBL_MAX;
            break;
         }
         add_datetime( &dt, i * 86400 );
         if( dt.year > 9999 ) {
            d = DBL_MAX + DBL_MAX;
            break;
         }
         d = dt.day + dt.month / 100.0 + dt.year / 1000000.0;
			break;
		case 22:
         if( !to_datetime( stack->stack[1], &dt ) ) return;              // Date difference
         i = dt.year;
         ts1 = calc_timestamp( &dt, &datetime_zero );
         if( !to_datetime( stack->stack[0], &dt ) ) return;
//         if( !check_days( ( dt.year - i ) * 366 ) ) return;
         if( abs( ( dt.year - i ) * 366 ) > INT_MAX / 86400 ) {
            d = DBL_MAX + DBL_MAX;
            break;
         }
         ts2 = calc_timestamp( &dt, &datetime_zero );
         d = (double)( ( ts1 - ts2 ) / 86400 );
			break;
		case 23:
			d = to_hms( from_hms(stack->stack[1]) + from_hms(stack->stack[0]) );   // HMS+
			break;
		default:
			d = to_hms( from_hms(stack->stack[1]) - from_hms(stack->stack[0]) );	// HMS-
			break;
	}
	if( !check_value(d) ) return;
	stack_drop();
	stack->stack[0] = d;
}

static double to_rad (double d) {
	switch (own_get_config (confid_angle_unit)) {
		case 0:	return d * M_PI / 180;
		case 2:	return d * M_PI / 200;
		default:	return d;
	}
}

static double from_rad (double d) {
	switch (own_get_config (confid_angle_unit)) {
		case 0:	return d * 180 / M_PI;
		case 2:	return d * 200 / M_PI;
		default:	return d;
	}
}

/* Returns the exchange rate of the built-in calculator
*/
static double exchange_rate (void) {

	eeprom_read (buf, eeprid_exchange_rate, EPM_NORMAL);
	return atof (buf);
}

/* Calculates a function (=one operand), function number
   256... in msg_argv[0]
*/
void calc_function (void) {

	double d;

	if( !stack_arg_check(1) ) return;

	switch(msg_argv[0]) {
		case 256:
			d = 1.0 / stack->stack[0];				// Inverse
			break;
		case 257:
			d = stack->stack[0] * stack->stack[0];		// Square
			break;
		case 258:
			d = sqrt(stack->stack[0]);           // Square root
			break;
		case 259:
			d = exp(stack->stack[0]);            // Exponent
			break;
		case 260:
			d = log (stack->stack[0]);				// Natural logarithm
			break;
		case 261:
			d = pow( 10.0, stack->stack[0] );		// Decimal exponent
			break;
		case 262:
			d = log(stack->stack[0]) / M_LN10;	// Decimal logarithm
			break;
		case 263:
         d = sinh(stack->stack[0]);           // Hyp. sine
			break;
		case 264:
			d = cosh(stack->stack[0]);           // Hyp. cosine
			break;
		case 265:
			d = tanh(stack->stack[0]);           // Hyp. tangent
			break;
		case 266:
			d = asinh(stack->stack[0]);	         // Hyp. arcus sine
			break;
		case 267:
			d = acosh(stack->stack[0]);          // Hyp. arcus cosine
			break;
		case 268:
			d = atanh(stack->stack[0]);          // Hyp. arcus tangent
			break;
		case 269:
			d = expm1(stack->stack[0]);          // Exponent-1
			break;
		case 270:
			d = log1p(stack->stack[0]);				// Natural logarithm(x+1)
			break;
		case 271:
			d = (double)~lround(stack->stack[0]);// Not
			break;
		case 272:
			d = sin (to_rad (stack->stack[0]));	// Sine
			break;
		case 273:
			d = cos (to_rad (stack->stack[0]));	// Cosine
			break;
		case 274:
			d = tan (to_rad (stack->stack[0]));	// Tangent
			break;
		case 275:
			d = from_rad (asin (stack->stack[0]));	// Arcus sine
			break;
		case 276:
			d = from_rad (acos (stack->stack[0]));	// Arcus cosine
			break;
		case 277:
			d = from_rad (atan (stack->stack[0]));	// Arcus tangent
			break;
		case 278:
			d = gamma1p(stack->stack[0]);        // !
			break;
		case 279:
			d = to_hms(stack->stack[0]);         // ->HMS
			break;
		case 280:
			d = from_hms(stack->stack[0]);       // HMS->
			break;
		case 281:
			d = stack->stack[0] * exchange_rate (); // To domestic
			break;
		default:
			d = stack->stack[0] / exchange_rate (); // To foreign
			break;
	}
	if( !check_value(d) ) return;
	stack_dup(0);
	stack->stack[0] = d;
}

/* Calculates a constant (=zero operands), constant number
   512... in message.argv[0]
*/
void calc_constant(void) {

	if(!stack_dup(1)) return;
	switch(msg_argv[0]) {
      default:                // Random number
         stack->stack[0] = ( ( rand() << 15 ) | rand() ) * ( 1.0 / ( 1 << 30 ) );
         break;
	}
}

/* Performs a stack operation, operation number
   in message.argv[0]
*/
void stack_operation (void) {

   unsigned int u;
   
	switch(msg_argv[0]) {
		case 0: 				// Pick
			if (!stack_dup (1)) return;
			stack->stack[0] = stack->stack[stack_ptr + 1];
			break;
		case 1:  				// Roll
			stack_roll (stack_ptr);
			break;
		case 2:   				// Rolld
			stack_rolld ();
			break;
		case 3:					// Rev
			stack_rev ();
			break;
		case 4:					// Dupn
			if (!stack_dup (stack_ptr + 1)) return;
			break;
		case 5:					// Dropn
			stack_drop();
			stack_ptr_check();
			break;
		case 6:					// Keep
			stack_set_depth( stack_ptr + 1 );
			break;
		default:             // Edit
         state |= STATE_EDIT;
         draw_stack();

			dtostr( stack->stack[stack_ptr], false );
         for( sign_ptr = u = 0; buf[u] != '\0'; ) {
				if (buf[u++] == 'E') sign_ptr = u;
			}
			edit_ptr = u;
			
         // In DEC mode, integers default to HEX
         // radix will be restored at draw_stack() when leaving edit mode
         if( radix < RADIX_16 ) radix = RADIX_16;

			draw_edit ();
			return;
	}
	draw_stack ();
}

static double calc_median(void) {

   double median = -( DBL_MAX + DBL_MAX ), min;
   unsigned int n, i, c;

   for( n = 0; n <= stack->depth / 2; ) {
      min = DBL_MAX + DBL_MAX;
      // Count the number of smallest value larger than <done> into c
      for( i = 0; i < stack->depth; i++ ) {
         if( stack->stack[i] <= median ) continue;
         if( stack->stack[i] < min ) {
            min = stack->stack[i];
            c = 1;
         } else if( stack->stack[i] == min ) {
            c++;
         }
      }
      // Even number of values and we were just about crossing the midpoint
      if( n == ( stack->depth + 1 ) / 2 ) return ( median + min ) / 2;
      
      median = min;
      n += c;
   }
   // Odd number of values or same value on both sides of the mid-point
   return median;
}

static double calc_mode(void) {

   double mode;
   unsigned int mode_n = 0, n, i, j;
   
   for( i = 0; i < stack->depth; i++ ) {
      // Count the number of d's in the rest of the stack->stack into n
      for( n = 1, j = i + 1; j < stack->depth; j++ ) {
         if( stack->stack[j] == stack->stack[i] ) n++;
      }
      if( n == mode_n ) {
         mode = DBL_MAX + DBL_MAX;  // mode found so far is not unique!
      } else if( n > mode_n ) {
         mode = stack->stack[i];           // new potential mode found
         mode_n = n;
      }
   }
   return mode;
}

/* Calculates a statistic (=N operands), statistic number
   768... in msg_argv[0]
*/
void calc_statistic(void) {

   unsigned int n;
   double x, delta, S, d;
   double stat[9];
#define  sum   stat[0]              // Total
#define  mean  stat[1]              // Mean
#define  sdev  stat[2]              // Standard deviation
#define  var   stat[3]              // Variance
#define  max   stat[4]              // Maximum
#define  min   stat[5]              // Minimum
#define  psdev stat[6]              // Population standard deviation
#define  pvar  stat[7]              // Population variance
#define  sum2  stat[8]              // Sum of x^2

	if( !stack_arg_check(1) ) return;

   switch(msg_argv[0]) {
      case 777:
         d = (double)stack->depth;   // Count
         break;
      case 778:
         d = calc_median();         // Median
         break;
      case 779:
         d = calc_mode();           // Mode
         break;
      default:
         // Robust variance calculation algorithm based on
         // Donald E. Knuth (1998). The Art of Computer Programming, volume 2:
         // Seminumerical Algorithms, 3rd edn., p. 232. Boston: Addison-Wesley.
         n = 0;
         mean = S = sum = sum2 = 0;
         max = -FLT_MAX;
         min = FLT_MAX;

         do {
            x = stack->stack[n++];

            sum += x;
            sum2 += x * x;
            if( x > max ) max = x;
            if( x < min ) min = x;

            delta = x - mean;
            mean += delta / n;
            S += delta * ( x - mean ); // This expression uses the new value of mean
         } while( n < stack->depth );
         var = S / ( n - 1 );
         sdev = sqrt(var);
         pvar = S / n;
         psdev = sqrt(pvar);
         d = stat[msg_argv[0] - 768];
         break;
   }

	if( !check_value(d) ) return;
   // Statistical functions do not consume their arguments
	if( !stack_dup(1) ) return;
	stack->stack[0] = d;

#undef   sum
#undef   mean
#undef   sdev
#undef   var
#undef   max
#undef   min
#undef   psdev
#undef   pvar
#undef   sum2
}


