/*
** stack.c
**
** Part of RPN calculator
** Copyright (c) 2005, 2006, 2007, 2010 NuukiaWorld
**
**  RPN calculator is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* includes nokix headers */
#include <nokix/all.h>

#include "main.h"
#include "draw.h"

/*******************************************************/

/*******************************************************/

/* Only these 4 bytes of RAM are statically reserved for this app.
*/
stack_t  *stack;

/*******************************************************/

/* Sets (stack_depth), allocates and frees the memory
*/
bool stack_set_depth( unsigned int depth ) {
   stack_t  *new_stack;

   // Discard previous last_stack if any
   if( last_stack != stack ) free(last_stack);
   last_stack = stack;

   if( depth == 0 ) {
      // Empty new stack
      stack = NULL;
      draw_action(); // Set NAVI2 to "Back"
   } else {
      // Non-empty new stack
      new_stack = malloc_try( sizeof(stack_t) + sizeof(double) * depth );
      if( new_stack == NULL ) {
         if( depth <= stack->alloc_depth ) { // No room to save stack
            // Continue to use current allocation because it is big enough
            display_msg( DMI_INFORMATION, &str_last_stack );
         } else {                            // Out of memory
            // Current allocation not big enough, abort
            display_msg( DMI_ERROR, &str_memory );
            return false;
         }
      } else {
         new_stack->alloc_depth = depth;
         // This may to illegal read access if (stack ==  NULL) or (depth > stack->depth)
         // However, it has no effect on this system without memory protection
         memcpy( new_stack->stack, stack->stack, sizeof(double) * depth );
         stack = new_stack;
      }
      stack->depth = depth;
   }
   return true;
}

/* Drops (stack_ptr + 1) bottommost values from stack
*/
void stack_drop(void) {
   unsigned int depth;
   
   depth = stack->depth - stack_ptr - 1;
   stack_set_depth(depth);
   memmove( stack->stack, last_stack->stack + stack_ptr + 1, sizeof(double) * depth );
}

/* Duplicates (count) bottommost values on stack
*/
bool stack_dup(unsigned int count) {
   unsigned int depth;

   depth = ( stack != NULL ? stack->depth : 0 ) + count;
	if( !stack_set_depth(depth) ) return false;
   memmove( stack->stack + count, stack->stack, sizeof(double) * ( depth - count ) );
   draw_action(); // Set NAVI2 to "Drop"
   return true;
}

/* Rolls stack at level (level + 1)
*/
void stack_roll( unsigned int level ) {
   double d;

   stack_set_depth( stack->depth );
	d = stack->stack[level];
	memmove( stack->stack + 1, stack->stack, sizeof(double) * level );
	stack->stack[0] = d;
}

/* Rolls stack down at level (stack_ptr + 1)
*/
void stack_rolld(void) {
	double d;

   stack_set_depth( stack->depth );
	d = stack->stack[0];
	memmove( stack->stack, stack->stack + 1, sizeof(double) * stack_ptr );
	stack->stack[stack_ptr] = d;
}

/* Reverses order of (stack_ptr + 1) bottommost values on stack->stack
*/
void stack_rev(void) {
	double d;
	unsigned int p1, p2;

   stack_set_depth( stack->depth );
	for (p1 = 0, p2 = stack_ptr; p1 < p2; p1++, p2--) {
		d = stack->stack[p2];
		stack->stack[p2] = stack->stack[p1];
		stack->stack[p1] = d;
	}
}

/* Checks that the stack contains at least (count) values
*/
bool stack_arg_check(unsigned int count) {

   if( stack == NULL || stack->depth < count ) {
      display_msg( DMI_INFORMATION, &str_arguments );
      return false;
   }
   return true;
}

/* Swaps current (stack) and (last_stack)
*/
void stack_undo(void) {
   stack_t *temp;
   
   temp = last_stack;
   last_stack = stack;
   stack = temp;
   draw_action();
}

/* Adjusts (stack_ptr) if it is outside the stack
*/
void stack_ptr_check(void) {
	if( stack == NULL ) {
		state &= ~STATE_STACK;   // Stack empty - exit stack mode
		stack_ptr = 0;
	} else if( stack_ptr >= stack->depth ) {
		stack_ptr = stack->depth - 1;
	}
}

